/*
 * Copyright (c) 2016, Pravin Nair <sreehari1390@gmail.com>
 * All rights reserved.
 *
 * This program is free software: you can use, modify and/or
 * redistribute it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later
 * version. You should have received a copy of this license along
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * @file fastbf.c
 * @brief fast bilateral filter and required functions
 *
 * @author PRAVIN NAIR  <sreehari1390@gmail.com>
 **/

#include "headersreq.h"

int shiftableBF(int m, int n, int sigmas, double sigmar, double** img, double** outimg, int cores, program_params* params,double eps);

/**
 * \brief Calculate norm of vector
 * \param a         Pointer to vector
 * \param n         Length of vector
 * \return Norm
 *
 * This routine calculates norm of vector
 * of length n.
 */
double norm(double *a,int n)
{
    double m_sum = a[0];
    int i;
    for ( i=1; i<n;++i){
        if (a[i]>m_sum) m_sum = a[i];
    }
    return m_sum;
}

/**
 * \brief Solve linear system Rx=b by back substitution
 * \param R         Pointer to upper triangular matrix R
 * \param b         Pointer to vector b
 * \param m         Length of vector b / Length of x
 * \return Pointer to solution vector x
 *
 * This routine solves the linear system Rx=b
 * (R is an upper triangular matrix) by back substitution.
 */
double *backsub(double **R, double *b , int m)
{
    double sum;
    int i,j;
    double *x=calloc(m,sizeof(double));
    for (i=m-1;i>=0;i--)
    {   
        sum=0;
        for (j=i;j<=m-1;j++)
        {
            sum=sum+(R[j][i]*(x[j]));
        }
        x[i]=(b[i]-sum)/R[i][i];
    }
    return x;
}

/**
 * \brief Apply fast shiftable bilateral filter to input image
 * \param m         Image height
 * \param n         Image width
 * \param sigmas    Standard deviation of spatial kernel
 * \param sigmar    Standard deviation of range kernel
 * \param img       Pointer to input image
 * \param outimg    Pointer to output image
 * \param cores     Number of physical cores on system
 * \param params    Pointer to Program parameters like
 *                  coefficients
 *
 * This routine applies the fast shiftable bilateral filter
 * with parameters sigmas & sigmar to input image img of
 * dimensions m x n and computes output image outimg.
 * The algorithm used is Fourier Basis approximation.
 * The Gaussian spatial convolutions are performed using
 * 'Deriche' or 'Young and van Vliet' fast recursive algorithms,
 * dpending on sigmar and max local dynamic range of the image.
 * The convolutions are performed parallelly with one thread
 * assigned for each physical core on the system.
 */
int shiftableBF(int m, int n, int sigmas, double sigmar, double** img, double** outimg, int cores, program_params* params,double eps) {
    int i,j,k,l;

    /** \brief Filter width */
    int w = 6*sigmas+1;
    /** \brief Filter radius */
    int c = (w-1)/2;
    
    /** \brief Matrix to store unnormalized filtered image */
    double **fnum=alloc_array(m,n);
    /** \brief Matrix to store weight sums for normalization */
    double **fdenum=alloc_array(m,n);
    double T=maxfilterfind(img,w,m,n);
    double Tmax = max(T,(3.2*sigmar));
    params->T = Tmax;
    /* Fourier Basis Algorithm */
    
    /* Period = 2*T+1 */   
    int K=(int)(2*T+1);
    double a1=(1/(2*T+1));
    double omegao=(2*M_PI)/(2*T+1);
    /* Vector holding DFT basis components in each iteration (for every frequency components)*/
    double* q=(double*)calloc(K,sizeof(double));
    /** \brief Range kernel discretization vector in range [-Tmax,Tmax]*/
    double* b=(double*)calloc(K,sizeof(double));
    double* recon=(double*)calloc(K,sizeof(double));
    int count=1;
    double* d=(double*)calloc(count,sizeof(double));    
    double* dmod;
    /* Variable holding coefficients in each iteration (for every frequency components)*/
    double temp=0;
    /* Calculating DC coefficient*/
    for (i=0;i<K;i++)
    {
        q[i]=1;
        b[i]=exp((-0.5*(i-Tmax)*(i-Tmax))/(sigmar*sigmar));
        temp += (a1*q[i]*b[i]);
    }
    double residualarr[K];
    for (i=0;i<K;i++)
    {
	recon[i]=temp*q[i];
        residualarr[i]=b[i]-recon[i];
    }   
    d[0]=temp;
    double residual=norm(residualarr,K);
    /* Calculating AC coefficients till range kernel approximation is less than eps*/
    while (residual>eps){
	temp=0;
	count=count+1;
        dmod=(double*)calloc(count,sizeof(double));     
        for (i=0;i<count-1;i++) dmod[i]=d[i];
        free(d);
	for (i=0;i<K;i++)
	{
            q[i]=cos((count-1)*omegao*(i-Tmax)); /* Calculating DFT basis */
            temp += (q[i]*b[i]);
        }        
	/* Multiplication by a1 is to normalize the basis and multiplying by 2 is to take into account for negative frequency
         component as well . Multiplying by 2 is done here so that we can avoid some checks in case of parallelization code
         as this is required only for AC frequency components. This is done to avoid changes in later part of code.*/
	temp*=(2*a1);
	for (i=0;i<K;i++)
        {
            recon[i]+=(temp*q[i]);
            residualarr[i]=b[i]-recon[i];
        }
        dmod[count-1]=temp;
        d=(double*)calloc(count,sizeof(double));
	for (i=0;i<count;i++) d[i]=dmod[i];
        free(dmod);
        residual=norm(residualarr,K);    
    }
    free(recon);
    free(q);
    free(b);
    params->N = count;
    params->coeff = (double*) calloc(count, sizeof(double));
    for (i=0; i<count; i++)  params->coeff[i] = d[i];
    /* End of algorithm for finding appropripriate number of DFT coefficients for range kernel approximation */
    /*******************************************************************************************************************/

    /** \brief Recursive parameter Ho to compute Auxiliary images */
    double complex** Ho = alloc_array_complex(m+w-1,n+w-1);    
    for (i=0;i<m;i++)
    {
        for (j=0;j<n;j++)
        {
            Ho[i+c][j+c]=exp(I*omegao*img[i][j]);
        }
    }
    #ifdef _OPENMP
        /* The auxiliary images are convolved with spatial Gaussian parallelly */
        int tid, procid;
        /** \brief Number of iterations assigned to each thread at fork */
        int chunk = (int) (count)/cores;
	if (chunk==0)
	    chunk=1; /* Auxillary image recursion is not required . So exp(I*omegao*k*img[i][j] has to be found for every k*/
        /* one thread per physical core */
        omp_set_num_threads(cores);
#pragma omp parallel shared(chunk,count,m,n,c,w,sigmas,fnum,fdenum,img,Ho,d,omegao) private(k,i,j,tid,procid)
    #endif
    {
        #ifdef _OPENMP
            tid = omp_get_thread_num();
            procid = sched_getcpu();
        #endif
        /** \brief Matrices for Auxiliary images */
        double complex **H=alloc_array_complex(m+w-1,n+w-1),**G=alloc_array_complex(m+w-1,n+w-1),**F=alloc_array_complex(m+w-1,n+w-1);
        #ifdef _OPENMP
            /** \brief fnum and fdenum private to thread */
            double ** fnum_p = alloc_array(m,n), **fdenum_p = alloc_array(m,n);
#pragma omp for schedule(static,chunk) nowait
        #endif
        for (k=0;k<count;k++)
        {
            /* Compute auxiliary images */
            #ifdef _OPENMP
                if (k%chunk==0) {
                    for (i=0;i<m;i++)
                    {
                        for (j=0;j<n;j++)
                        {
                            H[i+c][j+c]=exp(I*omegao*k*img[i][j]);
                            G[i+c][j+c]=conj(H[i+c][j+c]);
                            F[i+c][j+c]=G[i+c][j+c]*img[i][j];
                        }
                    }
            #else
                if (k==0) {
                    for (i=0;i<m;i++)
                    {
                        for (j=0;j<n;j++)
                        {
                            H[i+c][j+c]=1;
                            G[i+c][j+c]=1;
                            F[i+c][j+c]=G[i+c][j+c]*img[i][j];
                        }
                    }
            #endif
            } else {
                for (i=0;i<m;i++)
                {
                    for (j=0;j<n;j++)
                    {
                        H[i+c][j+c]=H[i+c][j+c]*Ho[i+c][j+c];
                        G[i+c][j+c]=conj(H[i+c][j+c]);
                        F[i+c][j+c]=G[i+c][j+c]*img[i][j];
                    }
                }
            }
    
            /* Gaussian filter applied to auxiliary images, algo decided by ratio T/sigmar */
            if ((Tmax/sigmar) < 3.5) {
                convolve_deriche2D(m,n,sigmas,F);  
                convolve_deriche2D(m,n,sigmas,G);

            } else {
                convolve_young2D(m,n,sigmas,F);    
                convolve_young2D(m,n,sigmas,G);
            }
            /* Update fnum and fdenum */
            #ifdef _OPENMP
                for (i=0;i<m;i++)
                {
                    for (j=0;j<n;j++)
                    {
                        fnum_p[i][j]+=crealf(d[k]*H[i+c][j+c]*F[i+c][j+c]);
                        fdenum_p[i][j]+=crealf(d[k]*H[i+c][j+c]*G[i+c][j+c]);
                    }
                }
        }

        /* Compute global fnum and fdenum from their private versions */
#pragma omp critical
        {
            for (i=0; i<m; i++) {
                for (j=0; j<n; j++) {
                    fnum[i][j] += fnum_p[i][j];
                    fdenum[i][j] += fdenum_p[i][j];
                }
            }
        }

        /* Deallocate thread private matrices */
        dealloc_array_fl(fnum_p,m);
        dealloc_array_fl(fdenum_p,m);
#else
            for (i=0;i<m;i++)
            {
                for (j=0;j<n;j++)
                {
                    fnum[i][j]+=crealf(d[k]*H[i+c][j+c]*F[i+c][j+c]);
                    fdenum[i][j]+=crealf(d[k]*H[i+c][j+c]*G[i+c][j+c]);
                }
            }
        }
#endif
        dealloc_array_fl_complex(H,m+w-1);
        dealloc_array_fl_complex(G,m+w-1);
        dealloc_array_fl_complex(F,m+w-1);
    }

    dealloc_array_fl_complex(Ho,m+w-1);

    /* Compute Output Image from fnum and fdenum */
    for (i=0;i<m;i++)
    {
        for (j=0;j<n;j++)
        {
            if (fabs(creal(fdenum[i][j]))<=0.001)
                outimg[i][j]=img[i][j];
            else 
                outimg[i][j]=(fnum[i][j]/fdenum[i][j]);
        }
    }
    
    dealloc_array_fl(fnum,m);
    dealloc_array_fl(fdenum,m);
    return EXIT_SUCCESS; 
}
