#include "headersreq.h"

/**
 * @file   fastbf_main.c
 * @brief  Main executable file.
 *
 * @author ANMOL POPLI  <anmol.ap020@gmail.com>
 *         PRAVIN NAIR  <sreehari1390@gmail.com>
 */

int main(int argc, char* argv[]) {
    int cores;
    #ifdef __linux__
        #ifdef _OPENMP
            cores = core_affinity();
        #endif
    #else
        #ifdef _OPENMP
            cores = omp_get_num_procs(); /** \brief Number of Hyperthreads for MP */
        #else
            cores=0;
        #endif
    #endif
    if (cores>10) cores=10;

    /* Check if there is a right call for algorithm */
    if (argc<=5){
        printf("Too less number of arguments \n");
        return EXIT_FAILURE;
    }
    program_params params; /** \brief Program parameters */

    /* Declaration parameters */
    double *input_image, *output_image;
    int columns,rows;
    struct timeval start,end;
    int i,j,sigmas;
    double sigmar,eps;

    /* Read image, image filename input as argv[1] */
    input_image = (double *)read_image(&columns, &rows, argv[1], IMAGEIO_DOUBLE | IMAGEIO_PLANAR | IMAGEIO_GRAYSCALE);

    sscanf(argv[2], "%d", &sigmas); /* SigmaS input as argv[2] */
    sscanf(argv[3], "%lf", &sigmar); /* SigmaR input as argv[3] */
    sscanf(argv[5], "%lf", &eps); /* parameter epsilon input as argv[4] */
    double** image = alloc_array(rows,columns); /** \brief Matrix to store input image */
    double** image_out = alloc_array(rows,columns); /** \brief Matrix to store output image */
    for (i=0; i<rows; i++) {
        for (j=0; j<columns; j++) {
            image[i][j] = input_image[i*columns+j]*255.0;
        }
    }
    free(input_image);

    gettimeofday(&start, NULL);
    /* Shiftable Bilateral Filter applied to image and result stored in image_out */
    if (shiftableBF(rows,columns,sigmas,sigmar,image,image_out,cores,&params,eps)!=EXIT_SUCCESS){
        printf("Fast bilateral filter algorithm failed \n");
        return EXIT_FAILURE;
    }
    gettimeofday(&end, NULL);

    printf("Execution time: %f ms\n", ((end.tv_sec-start.tv_sec)*1000.0) + ((end.tv_usec-start.tv_usec)/1000.0));
    output_image = (double*) calloc(rows*columns,sizeof(double));
    for (i=0; i<rows; i++) {
        for (j=0; j<columns; j++) {
            output_image[i*columns+j] = image_out[i][j]/255.0;
        }
    }
    /* Write image, image filename input as argv[4] */
    if (write_image(output_image, columns, rows, argv[4], IMAGEIO_DOUBLE | IMAGEIO_PLANAR | IMAGEIO_GRAYSCALE, 100)!=1){
        printf("Writing image failed \n");
        return EXIT_FAILURE;
    }
    dealloc_array_fl(image,rows);
    dealloc_array_fl(image_out,rows);
    free(output_image);
    #ifdef __linux__

        /** \brief Arrays to store range kernels for plotting */
        double* intensity = (double*) calloc(2*params.T+1,sizeof(double));
        double* rkernel = (double*) calloc(2*params.T+1,sizeof(double));
        double* rkernel_approx = (double*) calloc(2*params.T+1,sizeof(double));
        for (i=0; i<=2*params.T; i++) {
            intensity[i] = i-params.T;
            rkernel[i]  = exp(-0.5*intensity[i]*intensity[i]/(sigmar*sigmar));
        }

        /* Plot target and approximate range kernels */
        double T0 = max(params.T,(3.2*sigmar));
        double w0 = M_PI/T0;
        for (i=0; i<=2*params.T; i++) {
            rkernel_approx[i] = 0;
            for (j=0; j<params.N; j++) {
                rkernel_approx[i] += params.coeff[j]*cos(j*w0*intensity[i]);
            }
        }

        #ifndef IPOL
            gnuplot_ctrl    *   h1;
            h1 = gnuplot_init() ;
            gnuplot_cmd(h1, "set terminal png");
            gnuplot_cmd(h1,"set title \"Range Kernel Plot\"");
            gnuplot_setstyle(h1,"lines");
            gnuplot_cmd(h1, "set output \"rangekernel_plot.png\"");
            gnuplot_plot_xy(h1, intensity, rkernel, 2*params.T+1, "Target Gaussian");
            gnuplot_cmd(h1, "set output \"rangekernel_plot.png\"");
            gnuplot_plot_xy(h1, intensity, rkernel_approx, 2*params.T+1, "Fourier Approximation");
            gnuplot_close(h1);
        #else
            /* can't display plots on IPOL, so print to png images */

            gnuplot_ctrl *h1, *h2;
            h1 = gnuplot_init() ;
            gnuplot_cmd(h1, "set terminal pngcairo size 600,400 enhanced font \'Arial,10\'");
            gnuplot_cmd(h1, "set output \'RangeKernel_TargetGaussian.png\'");
            gnuplot_cmd(h1, "set title \"Range Kernel Plot\"");
            gnuplot_setstyle(h1, "lines") ;
            h2 = gnuplot_init() ;
            gnuplot_cmd(h2, "set terminal pngcairo size 600,400 enhanced font \'Arial,10\'");
            gnuplot_cmd(h2, "set output \'RangeKernel_FourierApprox.png\'");
            gnuplot_cmd(h2, "set title \"Range Kernel Plot\"");
            gnuplot_setstyle(h2, "lines") ;

            gnuplot_plot_xy(h1, intensity, rkernel, 2*params.T+1, "Target Gaussian");
            gnuplot_plot_xy(h2, intensity, rkernel_approx, 2*params.T+1, "Fourier Approximation");

            gnuplot_close(h1);
            gnuplot_close(h2);

        #endif
        free(intensity);
        free(rkernel);
        free(rkernel_approx);

    #endif // __linux__
    return EXIT_SUCCESS;
}
